/*
 * Interprocess communication message abstraction.
 *
 * Copyright (C) 2018, 2019, 2021, 2022, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/sys/ipc.h>
#include <l4/sys/utcb.h>
#include <l4/sys/types.h>

#include <sys/types.h>



EXTERN_C_BEGIN

/* Message abstraction. */

typedef struct
{
  /* Common virtual registers. */

  l4_buf_regs_t bregs;
  l4_msg_regs_t mregs;

  /* Input details. */

  l4_msgtag_t tag;
  unsigned int expected_items;

  /* Receive item counter. */

  unsigned int receive_items;

  /* Imported item counter. */

  unsigned int import_item;

  /* Output details. */

  unsigned int words;
  unsigned int items;

  /* Items transferred and not retained. */

  unsigned int discarded_items;
  l4_cap_idx_t to_discard[L4_UTCB_GENERIC_DATA_SIZE];

  /* Message label overriding. */

  l4_umword_t new_label;

  /* Server control. */

  int terminating;

} ipc_message_t;



/* Special status codes. */

#define IPC_MESSAGE_SENT      0x01d10000

/* Capability annotations. */

#define IPC_DISCARD_CAP_FLAG  (1 << (L4_CAP_SHIFT - 2))
#define discard_cap(x)        (x | IPC_DISCARD_CAP_FLAG)



/* Message operations. */

/* Lifecycle operations. */

void ipc_message_new(ipc_message_t *msg);
long ipc_message_expect(ipc_message_t *msg, unsigned int expected_items);
void ipc_message_request(ipc_message_t *msg, int op, l4_cap_idx_t endpoint);
void ipc_message_send(ipc_message_t *msg, int op, l4_cap_idx_t endpoint);
void ipc_message_wait(ipc_message_t *msg, l4_umword_t *label);
void ipc_message_reply(ipc_message_t *msg);
void ipc_message_discard(ipc_message_t *msg);
void ipc_message_free(ipc_message_t *msg);

/* Helper operations. */

void ipc_message_open(ipc_message_t *msg);
void ipc_message_prepare(ipc_message_t *msg);
void ipc_message_preserve_buffer_registers(ipc_message_t *msg);
void ipc_message_preserve_message_registers(ipc_message_t *msg);
void ipc_message_reset(ipc_message_t *msg);
void ipc_message_restore_buffer_registers(ipc_message_t *msg);
void ipc_message_restore_message_registers(ipc_message_t *msg);

/* Population operations. */

void ipc_message_add_capability(ipc_message_t *msg, l4_cap_idx_t cap);
void ipc_message_add_data(ipc_message_t *msg, const char *value, size_t length);
void ipc_message_add_item(ipc_message_t *msg, l4_cap_idx_t cap);
void ipc_message_add_fpage(ipc_message_t *msg, l4_snd_fpage_t fpage);
void ipc_message_add_page(ipc_message_t *msg, l4_umword_t hot_spot, l4_fpage_t fpage);
void ipc_message_add_receive_capability(ipc_message_t *msg, l4_cap_idx_t /* cap */);
void ipc_message_add_receive_fpage(ipc_message_t *msg, l4_snd_fpage_t fpage);
void ipc_message_add_string(ipc_message_t *msg, const char *value);
void ipc_message_add_word(ipc_message_t *msg, l4_umword_t value);
void ipc_message_propagate_item(ipc_message_t *msg, l4_cap_idx_t cap);
void *ipc_message_reserve_data(ipc_message_t *msg, size_t length);
void *ipc_message_reserve_words(ipc_message_t *msg, size_t length);
void ipc_message_send_error(ipc_message_t *msg, long error);

/* Access operations. */

l4_umword_t ipc_message_get_word(ipc_message_t *msg, unsigned int word);
l4_umword_t *ipc_message_get_word_address(ipc_message_t *msg, unsigned int word);
unsigned int ipc_message_number_of_items(ipc_message_t *msg);
unsigned int ipc_message_number_of_words(ipc_message_t *msg);

/* Supporting operations. */

void ipc_message_discard_capability(ipc_message_t *msg, l4_cap_idx_t cap);
void ipc_message_discard_dataspace(ipc_message_t *msg, l4_cap_idx_t mem, l4_addr_t addr);
long ipc_message_expect_capabilities(ipc_message_t *msg, int number);
long ipc_message_expect_capability(ipc_message_t *msg, int item);
void ipc_message_export_capability(ipc_message_t *msg, int item, l4_cap_idx_t ref);
void ipc_message_export_fpage(ipc_message_t *msg, int item, l4_snd_fpage_t fpage);
void ipc_message_export_page(ipc_message_t *msg, int item, l4_umword_t hot_spot, l4_fpage_t fpage);
long ipc_message_import_and_expect_capability(ipc_message_t *msg, int item, l4_cap_idx_t *ref);
long ipc_message_import_and_expect_dataspace(ipc_message_t *msg, int item, l4_cap_idx_t *mem, l4_addr_t *addr);
long ipc_message_import_capability(ipc_message_t *msg, int item, l4_cap_idx_t *ref);
long ipc_message_import_dataspace(ipc_message_t *msg, int item, l4_cap_idx_t *mem, l4_addr_t *addr);
long ipc_message_import_fpage(ipc_message_t *msg, int item, l4_snd_fpage_t *fpage);
void ipc_message_propagate_capability(ipc_message_t *msg, int item, l4_cap_idx_t ref);
void ipc_message_receive_fpage(ipc_message_t *msg, int item, l4_snd_fpage_t fpage);
l4_msgtag_t ipc_message_reply_tag(ipc_message_t *msg);
l4_msgtag_t ipc_message_request_tag(ipc_message_t *msg, int op);

EXTERN_C_END

/* vim: tabstop=2 expandtab shiftwidth=2
*/
